--
-- Data architecture
--
-- FuriousRaidDB = 
-- {
--		sets = 
--		{
--			['name'] = 
--			{
--				caste = 
--				enabled = 
--				monitor = 
--				{
--					['when'] =
--					{
--						item = 
--						{
--							id = 
--							name = 
--							link =
--						}
--						bids = 
--						{
--							['name'] =
--							{
--								tell =
--								value =
--							}
--						}
--					}
--				}
--				config = { }
--			}
--		}
--		players = 
--		{
--			['name'] =
--			{
--				class =
-- 				items =
--				{
--					['when'] =
--					{
--						set =
--						value =
--						item = 
--						{
--							id = 
--							name = 
--							link =
--						}
--					}
--				}
--				sets =
--				{
--					['name'] = 
--					{
--						<set specific data>
--					}
--				}
--			}
--		}
-- }

--
-- Set Definitions
--
-- definition = 
-- {
--		version = 
--		config = 
--		{
--		}
--		commands = 
--		{
--			['name'] = <function>
--		}
--	}
--
-- commands that the set should support
--	add(setname, playername, value) 
--		- this is the command that is used to add or remove
--		  hours from the player
--
--  clear(setname, playername)
--		- clears the player data
--
--  timer(setname)
--		- the timer callback for this set. starts/stops the timer.
--
--  timeractive(setname)
--		- returns the current time remaining on the timer or false if thereis none.
--
--	monitor(setname, monitorstate) [optional command -- can be nil]
--		- called everytime the monitor button is pressed. The set
--        communicates with FuriousRaid throught the monitorstate table.
--		  the monitor button will be grayed out if this command is nil.
--
--  monitortell(setname, monitorstate, playername, tell)
--		- called when the monitor is active and the the monitorstate
--		  is set to accept tells.
--
--	award(setname, playername, monitorstate)
--		- called after the item is MLed to the player, so the pertinent
--		  data can be stored
--
--	setvalue(setname)
--		- gets value to display in column C of the ui. Should be a
--		  simple text value
--
--	playervalue(setname, playername)
--		- gets players current raw set value
--
--  playervaluestring(setname, playername)
--		- gets a string representation of the player's set value
--
--	playerscore(setname, playername)
--		- gets a score based on the player's set value for sorting
--		  purposes.
--
--  showconfig(setname)
--		- display the configuration ui for the set.
--
--	print(setname, channel, [comparator], [destination])
--		- prints the player information to the designated channel
--
--



--
-- Monitor state
--
-- monitor = 
-- {
--		set = 
--		when =
--		stateid = 
--		monitortells =
--		enableaward =
--		buttontext =
--		item = 
--		{
--			id = 
--			name = 
--			link =
--		}
--		tells =
--		{
--			['name'] = 
--		}
-- }
--
-- stateid can be anything. however, to signal montioring is
-- finished all sets should use 'STOPPED' as the final state id

--
--
-- locals
local comm = ChatThrottleLib
local acecomm = LibStub('AceComm-3.0')
local aceserializer = LibStub('AceSerializer-3.0')

local fTip = CreateFrame('GameTooltip', 'FuriousRaidTooltip')
fTip:SetOwner(UIParent, 'ANCHOR_NONE')
fTip:AddFontStrings( fTip:CreateFontString( '$parentTextLeft1', nil, 'GameTooltipText'), fTip:CreateFontString( '$parentTextRight1', nil, 'GameTooltipText') )

--
--
-- globals
FuriousRaid = { prefix = 'FRaid', prefixsync = 'FRaidSync', prefixprint = 'FRaid:' }
FuriousRaidDB = { }
FuriousRaidDefault = { acceptbroadcast = true, ['version'] = 3.1, ['sets'] = { }, ['players'] = { }, ['waitlist'] = { }, }

FuriousRaid.state = {
	sets = { },
	filter = { 
		raid = false, 
		class = {
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PALADIN'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
	},
	raid = { },
	guild = { members = { }, ranks = { }, },
	selected = { },
	monitor = {
		set = 'UNKNOWN', 
		when = 'UNKNOWN',
		stateid = 'STOPPED',
		monitortells = false, 
		enableaward = false, 
		buttontext = 'Monitor: Off',
		item = {
			name = 'UNKNOWN',
		},
		tells = { },
	},
}

--
--
-- locals
local mainframe = CreateFrame('Frame', nil, UIParent)
local metafilters = { 
	-- { ['DEATHKNIGHT'] = true, ['DRUID'] = true, ['HUNTER'] = true, ['MAGE'] = true, ['PALADIN'] = true, ['PRIEST'] = true, ['ROGUE'] = true, ['SHAMAN'] = true, ['WARLOCK'] = true, ['WARRIOR'] = true, },
	['Armor'] = {
		['Miscellaneous'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PALADIN'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['Cloth'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PALADIN'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['Leather'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['PALADIN'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['Mail'] = { 
			['DEATHKNIGHT'] = true, 
			['HUNTER'] = true, 
			['PALADIN'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['Plate'] = { 
			['DEATHKNIGHT'] = true, 
			['PALADIN'] = true, 
			['WARRIOR'] = true, 
		},
		['Shields'] = { 
			['PALADIN'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['Librams'] = { 
			['PALADIN'] = true, 
		},
		['Idols'] = { 
			['DRUID'] = true, 
		},
		['Totems'] = { 
			['SHAMAN'] = true, 
		},
		['Sigils'] = { 
			['DEATHKNIGHT'] = true, 
		},
	},
	['Weapon'] = {
		['Bows'] = { 
			['HUNTER'] = true, 
			['ROGUE'] = true, 
			['WARRIOR'] = true, 
		},
		['Crossbows'] = { 
			['HUNTER'] = true, 
			['ROGUE'] = true, 
			['WARRIOR'] = true, 
		},
		['Daggers'] = { 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['Guns'] = { 
			['HUNTER'] = true, 
			['ROGUE'] = true, 
			['WARRIOR'] = true, 
		},
		['Fishing Poles'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PALADIN'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['Fist Weapons'] = { 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['Miscellaneous'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PALADIN'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['One-Handed Axes'] = { 
			['DEATHKNIGHT'] = true, 
			['HUNTER'] = true, 
			['PALADIN'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['One-Handed Maces'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['PALADIN'] = true, 
			['PRIEST'] = true, 
			['ROGUE'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['One-Handed Swords'] = { 
			['DEATHKNIGHT'] = true, 
			['HUNTER'] = true,
			['MAGE'] = true, 
			['PALADIN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['Polearms'] = { 
			['DRUID'] = true, 
			['DEATHKNIGHT'] = true, 
			['HUNTER'] = true, 
			['PALADIN'] = true, 
			['WARRIOR'] = true, 
		},
		['Staves'] = { 
			['DRUID'] = true, 
			['HUNTER'] = true, 
			['MAGE'] = true, 
			['PRIEST'] = true, 
			['SHAMAN'] = true, 
			['WARLOCK'] = true, 
			['WARRIOR'] = true, 
		},
		['Thrown'] = { 
			['HUNTER'] = true, 
			['ROGUE'] = true, 
			['WARRIOR'] = true, 
		},
		['Two-Handed Axes'] = { 
			['DEATHKNIGHT'] = true, 
			['HUNTER'] = true, 
			['PALADIN'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['Two-Handed Maces'] = { 
			['DEATHKNIGHT'] = true, 
			['DRUID'] = true, 
			['PALADIN'] = true, 
			['SHAMAN'] = true, 
			['WARRIOR'] = true, 
		},
		['Two-Handed Swords'] = { 
			['DEATHKNIGHT'] = true, 
			['HUNTER'] = true, 
			['PALADIN'] = true, 
			['WARRIOR'] = true, 
		},
		['Wands'] = { 
			['MAGE'] = true, 
			['PRIEST'] = true, 
			['WARLOCK'] = true, 
		},
	},
}

metafilters.User = {
	['MISCELLANEOUS'] = metafilters.Armor['Miscellaneous'],
	['CLOTH'] = metafilters.Armor['Cloth'],
	['LEATHER'] = metafilters.Armor['Leather'],
	['MAIL'] = metafilters.Armor['Mail'],
	['PLATE'] = metafilters.Armor['Plate'],
	['SHIELDS'] = metafilters.Armor['Shields'],
	['LIBRAMS'] = metafilters.Armor['Librams'],
	['IDOLS'] = metafilters.Armor['Idols'],
	['TOTEMS'] = metafilters.Armor['Totems'],
	['SIGILS'] = metafilters.Armor['Sigils'],
	['BOWS'] = metafilters.Weapon['Bow'],
	['CROSSBOWS'] = metafilters.Weapon['Crossbows'],
	['DAGGERS'] = metafilters.Weapon['Daggerd'],
	['GUNS'] = metafilters.Weapon['Guns'],
	['FISHING'] = metafilters.Weapon['Fishing Poles'],
	['FIST'] = metafilters.Weapon['Fist Weapons'],
	['1AXES'] = metafilters.Weapon['One-Handed Axes'],
	['1MACES'] = metafilters.Weapon['One-Handed Maces'],
	['1SWORDS'] = metafilters.Weapon['One-Handed Swords'],
	['POLEARMS'] = metafilters.Weapon['Polearms'],
	['STAVES'] = metafilters.Weapon['Staves'],
	['THROWN'] = metafilters.Weapon['Thrown'],
	['2AXES'] = metafilters.Weapon['Two-Handed Axes'],
	['2MACES'] = metafilters.Weapon['Two-Handed Maces'],
	['2SWORDS'] = metafilters.Weapon['Two-Handed Swords'],
	['WANDS'] = metafilters.Weapon['Wands'],
}

local temporarytable = { }

local deepcopy = function (_object)
	local lookup_table = { }

	local function _copy(_object)
		if type(_object) ~= 'table' then
			return _object
		elseif lookup_table[_object] then
			return lookup_table[_object]
		end

		local new_table = { }
		lookup_table[_object] = new_table
		for index, value in pairs(_object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(_object))
	end

	return _copy(_object)
end

local tablecount = function (_t)
	local count = 0
	for k, v in pairs(_t) do count = count + 1 end
	return count
end

--
--
-- sets
FuriousRaid.RegisterSetCaste = function (_caste, _definition)
	if _caste and _definition and type(_definition) == 'table' then 
		print(FuriousRaid.prefixprint, _caste .. ' registered')
		FuriousRaid.state.sets[_caste] = _definition 
	end
end

FuriousRaid.AddSet = function (_name, _caste)
	if FuriousRaid.state.sets[_caste] and not FuriousRaidDB.sets[_name] then
		FuriousRaidDB.sets[_name] = {
			name = _name,
			caste = _caste,
			monitors = { },
		}
		FuriousRaidDB.sets[_name].config = deepcopy(FuriousRaid.state.sets[_caste].config)
	end
end

FuriousRaid.RemoveSet = function (_name)
	FuriousRaidDB.sets[_name] = nil
end

FuriousRaid.GetCommands = function (_caste)
	if FuriousRaid.state.sets[_caste] then
		return FuriousRaid.state.sets[_caste].commands
	end
end

FuriousRaid.ExecuteCommand = function (_command, _name, ...)
	if FuriousRaidDB.sets[_name] and FuriousRaidDB.sets[_name].config and FuriousRaidDB.sets[_name].config[_command] and FuriousRaidDB.sets[_name].config[_command].broadcast then
		local t = { }

		table.insert(t, _command)
		table.insert(t, _name)
		for n = 1, select('#', ...) do table.insert(t,(select(n, ...) or '')) end

		acecomm:SendCommMessage(FuriousRaid.prefix, aceserializer:Serialize(t), 'GUILD', nil, 'BULK')
	end

	return FuriousRaid._ExecuteCommand(_command, _name, ...)
end
FuriousRaid._ExecuteCommand = function (_command, _name, ...)
	if FuriousRaidDB.sets[_name] and FuriousRaid.state.sets[FuriousRaidDB.sets[_name].caste] and FuriousRaid.state.sets[FuriousRaidDB.sets[_name].caste].commands[_command] then
		return FuriousRaid.state.sets[FuriousRaidDB.sets[_name].caste].commands[_command](_name, ...)
	else
		if not FuriousRaidDB.sets[_name] then print(FuriousRaid.prefixprint, 'ERROR EXECUTING ' .. _command .. ', set ' .. _name .. ' not found') 
		elseif not FuriousRaid.state.sets[FuriousRaidDB.sets[_name].caste] then print(FuriousRaid.prefixprint, 'ERROR EXECUTING ' .. _command .. ', caste ' .. FuriousRaidDB.sets[_name].caste .. ' not found')
		elseif not FuriousRaid.state.sets[FuriousRaidDB.sets[_name].caste].commands[_command] then print(FuriousRaid.prefixprint, 'ERROR EXECUTING ' .. _command .. ', command ' .. _command .. ' not found') end
	end
end

FuriousRaid.AddPlayer = function (_name, _class)
	if FuriousRaidDB.players[_name] == nil then
		print(FuriousRaid.prefixprint, 'Adding', _name, _class)
		FuriousRaidDB.players[_name] = {
			name = _name,
			class = _class,
			items = { },
			sets = { },
		}
	elseif FuriousRaidDB.players[_name].class == 'UNKNOWN' then
		print(FuriousRaid.prefixprint, 'Correcting class', _name, _class)
		FuriousRaidDB.players[_name].class = _class
	end
end

FuriousRaid.RemovePlayer = function (_name, _class)
	FuriousRaidDB.players[_name] = nil
end



--
--
-- filter
FuriousRaid.GetFilterString = function ()
	local f = ''
	
	if FuriousRaid.state.filter.raid then f = f .. 'raid' end
	
	for k, v in pairs(FuriousRaid.state.filter.class) do
		if v then f = f .. ' ' .. string.lower(k) end
	end
	
	return f
end

FuriousRaid.SetFilter = function (_raid, _group, _subgroup)
	FuriousRaid.ResetFilter()

	if type(_raid) == 'boolean' then FuriousRaid.state.filter.raid = _raid end

	if type(_group) == 'string' then
		if metafilters[_group] and metafilters[_group][_subgroup] then
			FuriousRaid.state.filter.class = metafilters[_group][_subgroup]
		end
	elseif type(_group) == 'table' then
		FuriousRaid.state.filter.class = _group
	end
end

FuriousRaid.SetFilterFromString = function (_filter)
	local raid = false
	_filter = string.upper(_filter)

	local x = { }
	for m in string.gmatch(_filter, '(%a+)') do 
		if metafilters.User[m] then
			for k, v in pairs(metafilters.User[m]) do x[k] = v end
		else
			x[m] = true 
		end
	end
	if x['RAID'] then raid = true x['RAID'] = nil end
	
	if tablecount(x) == 0 then 
		FuriousRaid.ResetFilter() 
		FuriousRaid.state.filter.raid = raid 
		return 
	end
	
	FuriousRaid.SetFilter(raid, x)
end

FuriousRaid.SetFilterFromItem = function (_raid, _itemlink)
	local x, m

	fTip:SetOwner(UIParent, 'ANCHOR_NONE')
	fTip:ClearLines()
	fTip:SetHyperlink(_itemlink)
	for i = 2, fTip:NumLines() do
		local t = _G['FuriousRaidTooltipTextLeft' .. i]:GetText()
		local _, _, c = string.find(t, 'Classes: (.+)')

		if c then
			c = string.upper(string.gsub(string.gsub(c, ' ', ''), ',', ' '))

			local x = { }
			for m in string.gmatch(c, '(%a+)') do
				x[m] = true
			end

			FuriousRaid.SetFilter(_raid, x)
			fTip:Hide()
			return
		end
	end
	fTip:Hide()

	if not x then
		local _, _, _, _, _, iType, iSubType, _, _, _ = GetItemInfo(_itemlink)
		FuriousRaid.SetFilter(_raid, iType, iSubType)
	end
end

FuriousRaid.ResetFilter = function ()
	FuriousRaid.state.filter.raid = false
	FuriousRaid.state.filter.class = {
		['DEATHKNIGHT'] = true, 
		['DRUID'] = true, 
		['HUNTER'] = true, 
		['MAGE'] = true, 
		['PALADIN'] = true, 
		['PRIEST'] = true, 
		['ROGUE'] = true, 
		['SHAMAN'] = true, 
		['WARLOCK'] = true, 
		['WARRIOR'] = true, 
	}
end

FuriousRaid.IsFiltered = function (_player)
	local r = (not FuriousRaid.state.filter.raid) or FuriousRaid.state.raid[_player]
	local c = FuriousRaidDB.players[_player] and FuriousRaid.state.filter.class[FuriousRaidDB.players[_player].class]

	return r and c
end

FuriousRaid.GetFilteredList = function ()
	temporarytable = { }
 
	for k, v in pairs(FuriousRaidDB.players) do
		if FuriousRaid.IsFiltered(k) then table.insert(temporarytable, k) end
	end

	return temporarytable
end

FuriousRaid.GetRaidList = function ()
	return FuriousRaid.state.raid
end

FuriousRaid.GetWaitList = function ()
	return FuriousRaidDB.waitlist
end

FuriousRaid.ResetMonitorState = function ()
	FuriousRaid.state.monitor = {
		set = FuriousRaid.state.selected.set or 'UNKNOWN',
		when = time(),
		stateid = 'STOPPED',
		monitortells = false, 
		enableaward = false, 
		buttontext = 'Monitor: Off',
		item = {
			name = 'UNKNOWN',
		},
		tells = { },
	}
end

FuriousRaid.IsMasterLootable = function (_player)
	for i = 1, 40 do
		if GetMasterLootCandidate(i) == _player then return true end
	end

	return false
end
StaticPopupDialogs['FURIOUSRAID_ML'] = {
	text = CONFIRM_LOOT_DISTRIBUTION,
	button1 = YES,
	button2 = NO,
	OnAccept = function(f, d)
		GiveMasterLoot(d.slot, d.player)
	end,
	timeout = 0,
	hideOnEscape = 1,
}
FuriousRaid.MasterLoot = function (_player, _id)
	local looticon, lootname, lootquantity, lootrarity, lootlink, lootid
	local lootslotindex = 0
	local lootplayerindex = 0

	for i = 1, 40 do
		if GetMasterLootCandidate(i) == _player then
			lootplayerindex = i
			break
		end
	end

	local lf = LootFrame
	if XLootFrame then lf = XLootFrame end

	if lf:IsVisible() then
		for lsi = 1, GetNumLootItems() do
			looticon, lootname, lootquantity, lootrarity = GetLootSlotInfo(lsi)

			if lootquantity > 0 then
				lootlink = GetLootSlotLink(lsi) 
				_, _, lootid = string.find(lootlink, 'item:([0-9]+):')
				if tonumber(lootid) == _id then
					lootslotindex = lsi
					break
				end
			end
		end
	end

	if lootplayerindex > 0 and lootslotindex > 0 then
		local popup = StaticPopup_Show('FURIOUSRAID_ML', ITEM_QUALITY_COLORS[lootrarity].hex .. lootname .. FONT_COLOR_CODE_CLOSE, _player)
		if popup then popup.data = { slot = lootslotindex, player = lootplayerindex } end	
		--GiveMasterLoot(lootslotindex, lootplayerindex)
		return true
	end

	return false
end

FuriousRaid.IsTimerActive = function ()
	for k, v in pairs(FuriousRaidDB.sets) do
		if v.config.timer.handle then return true end
	end
	
	return false
end
 
FuriousRaid.DataIntegrityCheck = function ()
	for k, v in pairs(FuriousRaidDB.players) do
		for kk, vv in pairs(v.sets) do
			if not FuriousRaidDB.sets[kk] or FuriousRaid.ExecuteCommand('playervalue', kk, k) == 0 then
				FuriousRaidDB.players[k].sets[kk] = nil
			end
		end
	end

	for k, v in pairs(FuriousRaidDB.players) do
		if not v.sets or tablecount(v.sets) == 0 then
			print(FuriousRaid.prefixprint, 'Bad player: ' .. k)
			FuriousRaidDB.players[k] = nil
		end
	end

	FuriousRaid.UpdateRaidList()
end

FuriousRaid.DataIntegrityCheckSpecial = function ()
	-- fix up my broken set monitors :P
	for k, v in pairs(FuriousRaidDB.sets) do
		for kk, vv in pairs(v.monitors) do
			local newbids = { }
			for kkk, vvv in pairs(vv.bids) do
				if tonumber(kkk) then
					table.insert(newbids, vvv)
				else
					if vvv.tell then
						vvv.name = kkk
						vvv.score = vvv.value	
						table.insert(newbids, vvv)
					else
						if FuriousRaidDB.players[kkk] and FuriousRaidDB.players[kkk].items[kk] then
							vv.item = FuriousRaidDB.players[kkk].items[kk].item
						end
					end
				end
			end
			vv.bids = newbids
		end
	end
end

FuriousRaid.UpdateRaidList = function ()
	local count = GetNumRaidMembers()
	if count == 0 then 
		FuriousRaid.state.raid = { } 
		return 
	end
	
	for k, v in pairs(FuriousRaid.state.raid) do v.updated = nil end
	
	for i = 1, count do
		local nm, rk, sg, lv, cl, fn, zn, ol, id, rl, ml = GetRaidRosterInfo(i)
		FuriousRaid.state.raid[nm] = {
			name = nm,
			rank = rk,
			group = sg,
			level = lv,
			class = fn,
			zone = zn,
			online = ol,
			dead = id,
			role = rl,
			masterlooter = ml,
			updated = true, 
		}

		if FuriousRaid.IsTimerActive() then
			FuriousRaid.AddPlayer(nm, fn)
		end

		if FuriousRaidDB.waitlist[nm] then
			FuriousRaidDB.waitlist[nm] = nil
			FuriousUI.UpdateWaitlist()
		end
	end
	
	for k, v in pairs(FuriousRaid.state.raid) do 
		if not v.updated then FuriousRaid.state.raid[k] = nil end
	end	
end

FuriousRaid.UpdateGuildList = function ()
	if not IsInGuild()then 
		FuriousRaid.state.guild.members = { } 
		return 
	end
	
	for k, v in pairs(FuriousRaid.state.guild.members) do v.updated = nil end
	
	local count = GetNumGuildMembers(true)
	for i = 1, count do
		local nm, rk, ri, lv, _, zn, no, ono, ol, st, fn = GetGuildRosterInfo(i)
		if nm then
			FuriousRaid.state.guild.members[nm] = {
				name = nm,
				rank = rk,
				rankindex = ri,
				level = lv,
				class = fn,
				zone = zn,
				online = ol,
				note = no,
				officernote = ono,
				status = st,
				online = ol,
				updated = true, 
			}
		end
	end
	
	for k, v in pairs(FuriousRaid.state.guild.members) do 
		if not v.updated then FuriousRaid.state.guild.members[k] = nil end
	end	
	
	count = GuildControlGetNumRanks()
	for i = 1, count do FuriousRaid.state.guild.ranks[i] = GuildControlGetRankName(i) end
	for i = (count+1), #(FuriousRaid.state.guild.ranks) do FuriousRaid.state.guild.ranks[i] = nil end
end

FuriousRaid.CommEventBroadcast = function (_prefix, _message, _distribution, _sender)
	if _sender == UnitName('player') then return end
	if not FuriousRaidDB.acceptbroadcast then return end
	
	local s, t = aceserializer:Deserialize(_message)
	if s then 
		FuriousRaid._ExecuteCommand(unpack(t)) 
	else
		print(FuriousRaid.prefixprint, t)
	end
end
acecomm:RegisterComm(FuriousRaid.prefix, FuriousRaid.CommEventBroadcast)

StaticPopupDialogs['FURIOUSRAID_SYNC'] = {
	text = 'Furious|cff00ff00Raid|r: Accept Sync from %s?',
	button1 = 'Accept',
	button2 = 'Decline',
	OnAccept = function (f, d)
		print(FuriousRaid.prefixprint, 'Sync accepted -- deserializing')	
		local s, t = aceserializer:Deserialize(d)
		if s then 
			print(FuriousRaid.prefixprint, 'Deserialization complete -- updating database')
			FuriousRaid.state.selected = { }
			FuriousUI.Hide()
			FuriousRaidDB = t
		else
			print(FuriousRaid.prefixprint, t)
		end
	end,
	OnCancel = function()
		print(FuriousRaid.prefixprint, 'Sync declined')
	end,
	timeout = 90,
	whileDead = 1,
	hideOnEscape = 0,
}
FuriousRaid.CommEventSync = function (_prefix, _message, _distribution, _sender)
	if _sender == UnitName('player') then return end
	
	local popup = StaticPopup_Show('FURIOUSRAID_SYNC', _sender)
	if popup then popup.data = _message end	
end
acecomm:RegisterComm(FuriousRaid.prefixsync, FuriousRaid.CommEventSync)

FuriousRaid.CommSendSync = function (_player)
	print(FuriousRaid.prefixprint, 'Serializing database')
	local t = aceserializer:Serialize(FuriousRaidDB)
	print(FuriousRaid.prefixprint, 'Serialization complete -- sending')
	acecomm:SendCommMessage(FuriousRaid.prefixsync, t, 'WHISPER', _player, 'BULK')
end


--
--
-- updates
local lastupdate = 0
mainframe.FUpdates = function (f, e)
	lastupdate = lastupdate + e

	if (lastupdate > 1) then
		lastupdate = 0
		
		for k, v in pairs(FuriousRaidDB.sets) do
			v.config.timer.remaining = FuriousRaid.ExecuteCommand('timeractive', k)
		end
	end
end


--
--
-- event handling
local eventtable = {}
mainframe.FEvents = function (f, e, ...)
	if e == 'VARIABLES_LOADED' then
		if FuriousRaidDB.account then
			print(FuriousRaid.prefixprint, 'WARNING. converting DB to new version')

			local newDB = { version = 3.1, sets = { }, players = { }, waitlist = { }, }

			for k, v in pairs(FuriousRaidDB.account.Set) do
				newDB.sets[k] = {
					enabled = true,
					caste = v.Type,
					monitors = { },
					config = { }, 
				}
			end

			for k, v in pairs(FuriousRaidDB.account.Karma) do
				newDB.players[k] = {
					class = string.upper(v.class) or 'UNKOWN',
					sets = v.karma,
					items = { },
				}
			end

			local t = { year = 0, month = 0, day = 0, hour = 0, min = 0, }
			for k, v in pairs(FuriousRaidDB.account.Monitors) do
				if v.item.id ~= 'NOITEM' then
					_, _, t.year, t.month, t.day, t.hour, t.min = string.find(k, '(%d+)-(%d+)-(%d+)%s+(%d+):(%d+)')
					local w = time(t)
					if newDB.sets[v.set] then
						newDB.sets[v.set].monitors[w] = {
							item = {
								id = tonumber(v.item.id),
								name = v.item.name,
								link = v.item.link,
							},
							bids = { },
						}

						for kk, vv in pairs(v.bids) do
							for kkk, vvv in pairs(vv) do
								if vvv.name then
									table.insert(newDB.sets[v.set].monitors[w].bids, {
										name = vvv.name,
										tell = vvv.rawdata,
										value = vvv.value,
										score = vvv.value,
									})

									if vvv.item then
										newDB.sets[v.set].monitors[w].awarded = vvv.name

										if newDB.players[vvv.name] then
											newDB.players[vvv.name].items[w] = {
												item = {
													id = tonumber(v.item.id),
													name = v.item.name,
													link = v.item.link,
												},
												value = vvv.value,
												set = v.set,
											}
										end
									end
								end
							end
						end
					end
				end
			end

			FuriousRaidDB = newDB
			print(FuriousRaid.prefixprint, 'WARNING. DB updated please reload your UI')
		end

		if FuriousRaidDB.acceptbroadcast == nil then FuriousRaidDB.acceptbroadcast = true end

		if not FuriousRaidDB.version or FuriousRaidDB.version ~= FuriousRaidDefault.version then
			print(FuriousRaid.prefixprint, 'Incorrect savedvairable version reverting to defaults')
			FuriousRaidDB = FuriousRaidDefault
		end

		-- verify set config
		for k, v in pairs(FuriousRaidDB.sets) do
			v.enabled = (FuriousRaid.state.sets[v.caste] ~= nil)

			if v.enabled then
				if not v.config or not v.config.version or v.config.version ~= FuriousRaid.state.sets[v.caste].config.version then
					print(FuriousRaid.prefixprint, 'Invalid config in ' .. k .. ' reverting to default')
					v.config = deepcopy(FuriousRaid.state.sets[v.caste].config)
				end

				v.config.timer.handle = nil
			end
		end
		
		GuildRoster()
		FuriousRaid.UpdateRaidList()

		mainframe:SetScript('OnUpdate', mainframe.FUpdates)
	end

	if e == 'RAID_ROSTER_UPDATE' then
		FuriousRaid.UpdateRaidList()
	end

	if e == 'GUILD_ROSTER_UPDATE' then
		FuriousRaid.UpdateGuildList()
	end

	if e == 'PLAYER_ENTERING_WORLD' then
		FuriousRaid.UpdateRaidList()
	end

	if e == 'CHAT_MSG_WHISPER' then
		local name = arg2
		local tell = string.lower(arg1)

		if string.find(tell, string.lower(FuriousRaid.prefix)) then return end

		if FuriousRaid.state.monitor.monitortells then
			FuriousRaid.state.monitor.tells[name] = tell
			FuriousRaid.ExecuteCommand('monitortell', FuriousRaid.state.monitor.set, FuriousRaid.state.monitor, name, tell)
			return
		end

		-- check for you
		if name == UnitName('player') then return end

		-- check for tell to use for monitor
		if string.find(tell, 'item:([0-9]+):') then
			_, _, lootid = string.find(tell, 'item:([0-9]+):')			
			FuriousRaid.state.whisperloot = tonumber(lootid)
		end

		-- check for WL tell
		if string.find(tell, 'wl') then
			if FuriousRaid.state.raid[name] then
				comm:SendChatMessage('BULK', FuriousRaid.prefix, FuriousRaid.prefix .. ' You are already in the raid', 'WHISPER', nil, name)
			else
				comm:SendChatMessage('BULK', FuriousRaid.prefix, FuriousRaid.prefix .. ' You have been placed on the waitlist', 'WHISPER', nil, name)
				if FuriousRaid.state.guild.members[name] then
					FuriousRaid.AddPlayer(name, FuriousRaid.state.guild.members[name].class)
				else
					FuriousRaid.AddPlayer(name, 'UNKNOWN')
				end
				
				FuriousRaidDB.waitlist[name] = GetTime()
				FuriousUI.UpdateWaitlist()
			end
		end
	end

	if e == 'CHAT_MSG_ADDON' then
		if arg1 == FuriousRaid.prefix and arg4 ~= UnitName('player') then
			t = {}
			for v in string.gmatch(arg2, '([^,]+)') do table.insert(t, tonumber(v) or v) end
			FuriousRaid._ExecuteCommand(unpack(t))
		end
	end
end
mainframe:SetScript('OnEvent', mainframe.FEvents)
mainframe:RegisterEvent('VARIABLES_LOADED')
mainframe:RegisterEvent('RAID_ROSTER_UPDATE')
mainframe:RegisterEvent('GUILD_ROSTER_UPDATE')
mainframe:RegisterEvent('CHAT_MSG_WHISPER')
mainframe:RegisterEvent('PLAYER_ENTERING_WORLD')


--
--
-- BigWigs / DBM
local raidmod
if BigWigs then 
	raidmod = BigWigs:New('FuriousRaid', 3.1)
	raidmod._type = 'BigWigs'
	raidmod._bars = { }
elseif DBM then
	raidmod = DBM:NewMod('FuriousRaid', 'FuriousRaid')
	raidmod._type = 'DBM'
	raidmod._bars = { }
	raidmod._localization = DBM:GetModLocalization('FuriousRaid')
end

FuriousRaid.StartBar = function (_id, _seconds, _text, _icon)
	if not raidmod then return end
	
	if not raidmod._bars[_id] then 
		if raidmod._type == 'BigWigs' then
			raidmod._bars[_id] = _text
		elseif raidmod._type == 'DBM' then
			raidmod._localization.timers[_id] = _text
			raidmod._bars[_id] = raidmod:NewTimer(_seconds, _id, _icon)
		end
	end
	
	if raidmod._type == 'BigWigs' then
		raidmod:TriggerEvent('BigWigs_StartBar', raidmod, _text, _seconds, _icon)
	elseif raidmod._type == 'DBM' then
		raidmod._bars[_id]:Start()
	end
end

FuriousRaid.StopBar = function (_id)
	if not raidmod then return end

	if not raidmod._bars[_id] then return end
	
	if raidmod._type == 'BigWigs' then
		raidmod:TriggerEvent('BigWigs_StopBar', raidmod, raidmod._bars[_id])
	elseif raidmod._type == 'DBM' then
		raidmod._bars[_id]:Stop()
	end
end
